
#include "ChargeMapCreator.H"
#include "Map.H"
#include "PlayerInfo.H"
#include "StaticMap.H"
#include "Blackboard.H"
#include "TBConfig.H"
#include "Options.H"
#include "EventTimer.H"

enum { TILE_UNKNOWN=0, TILE_WATER=1, TILE_GROUND=2, TILE_CLIFF=3 };

using namespace std;

bool is_initiated = false;
bool cm_thread_running = false;

/*
	Constructor.
*/
ChargeMapCreator::ChargeMapCreator() {
	
}

/*
	Destructor.
*/
ChargeMapCreator::~ChargeMapCreator() {
	
}

/*
	Returns true if the chargemap is ready for use, i.e. terrain and base charges are generated.
*/
bool ChargeMapCreator::isReady() {
	if (is_initiated && !cm_thread_running) {
		return true;
	}

	return false;
}

/*
	Returns false if the charge map needs to be initiated, i.e. the terrain and base
	charges must be generated.
*/
bool ChargeMapCreator::isInitiated() {
	if (!is_initiated && !cm_thread_running) {
		return false;
	}
	return true;
}

/*
	Returns a pointer to the charge map.
*/
ChargeMap* ChargeMapCreator::getMap() {
	return c_map;
}

/*
	Initiates the map. This method is threaded since it usually takes about 150 ms and each action
	frame is 125 ms.
*/
void ChargeMapCreator::initMap(GameStateModulePtr mState) {
	if (cm_thread_running) return;

	cm_thread_running = true;

	state = mState;

	playerId = state->get_game().get_client_player();
	enemyId = 0;
	if (playerId == 0) enemyId = 1;

	c_map = new ChargeMap(state);

	obj_map = new ObjMap();
	
	Options::get("-debug", debug);
	
	//Starts the SDL thread
	my_thread = SDL_CreateThread(ChargeMapCreator::execute, this);
}

/*
	Init thread run method.
*/
int ChargeMapCreator::execute(void *data) {
	ChargeMapCreator &cm = *static_cast<ChargeMapCreator*>(data);

	cout << "Starting map init thread" << endl;

	const Game& game(cm.state->get_game());

	clock_t start = clock();
	cout << "Creating terrain ChargeMap...";
	clock_t tStart = clock();
	cm.createTerrainMap(game);
	cm.postProcessTerrain2();
	cm.postProcessTerrain();
	//cm.finalPostProcessTerrain();
	clock_t tEnd = clock();
	cout << "done in " << (tEnd - tStart) << " ms" << endl;

	cout << "Creating base objects ChargeMap...";
	clock_t bStart = clock();
	cm.updateBaseMap(game);
	clock_t bEnd = clock();
	cout << "done in " << (bEnd - bStart) << " ms" << endl;
	
	clock_t end = clock();
	cout << "Map creation took " << (end - start) << " ms" << endl;

	is_initiated = true;

	cm_thread_running = false;

	cout << "Stopping map init thread" << endl;

	return 0;
}

/*
	Returns the ObjMap.
*/
ObjMap* ChargeMapCreator::getObjMap() {
	return obj_map;
}

/*
	Updates own-, enemy- and neutral units in the ObjMap.
*/
void ChargeMapCreator::updateObjMap(const Game& game) {
	obj_map->setAllDestroyed();

	const sint4 my_id = game.get_client_player();
	const sint4 en_id = 1 - my_id;
	
	const Game::ObjCont &objs = game.get_objs(my_id);
	FORALL (objs, it) {
		GameObj* gob = (*it)->get_GameObj();
		if (gob && gob->sod.in_game && *gob->sod.x >= 0 && !gob->is_dead()) {
			std::string objType = gob->bp_name();
			if (objType == "marine") {
				uint4 unitId = game.get_cplayer_info().get_id(gob);
				obj_map->updateUnitPosition(game.tick(), unitId, Blackboard::TYPE_MARINE, *gob->sod.x, *gob->sod.y);
			}
			if (objType == "tank") {
				uint4 unitId = game.get_cplayer_info().get_id(gob);
				obj_map->updateUnitPosition(game.tick(), unitId, Blackboard::TYPE_TANK, *gob->sod.x, *gob->sod.y);
			}
		}
	}

	const Game::ObjCont &eobjs = game.get_objs(en_id);
	FORALL (eobjs, it) {
		GameObj* gob = (*it)->get_GameObj();
		if (gob && gob->sod.in_game && *gob->sod.x >= 0 && !gob->is_dead()) {
			std::string objType = gob->bp_name();
			if (objType == "marine") {
				uint4 unitId = game.get_cplayer_info().get_id(gob);
				obj_map->updateEnemyUnitPosition(game.tick(), unitId, Blackboard::TYPE_MARINE, *gob->sod.x, *gob->sod.y);
			}
			if (objType == "tank") {
				uint4 unitId = game.get_cplayer_info().get_id(gob);
				obj_map->updateEnemyUnitPosition(game.tick(), unitId, Blackboard::TYPE_TANK, *gob->sod.x, *gob->sod.y);
			}
		}
	}

	int n = game.get_player_num();
	const Game::ObjCont &nobjs = game.get_objs(n);
	FORALL (nobjs, it) {
		GameObj* gob = (*it)->get_GameObj();
		if (gob && gob->sod.in_game && *gob->sod.x >= 0 && !gob->is_dead()) {
			std::string objType = gob->bp_name();
			if (objType == "sheep") {
				uint4 sheepId = game.get_cplayer_info().get_id(gob);
				obj_map->updateSheepPosition(game.tick(), sheepId, *gob->sod.x, *gob->sod.y);
			}
		}
	}
}

/*
	Updates the unit charge map to the current game state. Shall be called
	each action frame.
*/
void ChargeMapCreator::updateUnitMap(const Game& game) {
	updateObjMap(game);
}

/*
	Updates the base charge map to the current game state. Shall be called
	when a enemy or friendly base is destroyed.
*/
void ChargeMapCreator::updateBaseMap(const Game& game) {
	obj_map->clearBaseLists();
	updateObjBaseMap(game);
}

/*
	Updates own- and enemy bases in the ObjMap.
*/
void ChargeMapCreator::updateObjBaseMap(const Game& game) {
	int n = game.get_player_num();
	PlayerInfo& pi = game.get_cplayer_info();
	const sint4 cid = game.get_client_player();
	
	bool found = false;
	//Iterate through all players. The last player (n) is the server objects, but we dont
	//need to change charge for neutral server objects.
	for (int player = 0; player < n; player++) {
		const Game::ObjCont &objs = game.get_objs(player);

		bool friendly = false;
		if (player == cid) {
			friendly = true;
		}

		//Iterate through all objects for this player
		FORALL (objs, it) {
			GameObjPtr gob = (*it)->get_GameObj();
			if (gob && gob->sod.in_game && *gob->sod.x >= 0 && !gob->is_dead()) {
				//Valid object found
				uint4 objId = pi.get_id(gob);
				
				std::string objType = gob->bp_name();
				
				if (objType == "controlCenter") {
					if (friendly) {
						found = true; //Base found
						obj_map->addBase(objId, (*gob->sod.x), (*gob->sod.y), (*gob->sod.x1), (*gob->sod.y1), (*gob->sod.x2), (*gob->sod.y2), true);
					}
					else {
						obj_map->addBase(objId, (*gob->sod.x), (*gob->sod.y), (*gob->sod.x1), (*gob->sod.y1), (*gob->sod.x2), (*gob->sod.y2), false);
					}
				}
			}
		}
	}

	if (!found) {
		//No bases found. Lock state to hunt units
		Blackboard::lockState(Blackboard::STATE_HUNT_UNITS);
	}
}

/*
	Creates charge map for the terrain.
*/
void ChargeMapCreator::createTerrainMap(const Game &game) {
	const Map<GameTile> &gameMap = game.get_map();

	//Iterate through the map tiles
	FORS (y, gameMap.get_height()) {
		FORS (x, gameMap.get_width()) {
			const GameTile &tile = gameMap(x,y);

			//Start and end points for this tile in our charge map
			int cmTilesPerGameTile = game.get_tile_points() / ChargeMap::CMTILE_WIDTH;

			int startX = x * cmTilesPerGameTile;
			int endX = startX + cmTilesPerGameTile;
			int startY = y * cmTilesPerGameTile;
			int endY = startY + cmTilesPerGameTile;

			if (tile.get_split() == Tile::NO_SPLIT) {
				//No split. Tile only has one terrain type

				if (tile.get_typeW() != TILE_GROUND) {
					for (int cX = startX; cX < endX; cX++) {
						for (int cY = startY; cY < endY; cY++) {
							handleImpassableTerrainTile(cX, cY);
						}
					}
				}
			}
			else if (tile.get_split() == Tile::BT_SPLIT) {
				//Tile is split |\|
				if (tile.get_typeW() != TILE_GROUND) {
					for (int cX = startX; cX < endX; cX++) {
						for (int cY = startY; cY < endY; cY++) {
							if (y >= x) {
								handleImpassableTerrainTile(cX, cY);
							}
						}
					}
				}
				else if (tile.get_typeE() != TILE_GROUND) {
					for (int cX = startX; cX < endX; cX++) {
						for (int cY = startY; cY < endY; cY++) {
							if (y <= x) {
								handleImpassableTerrainTile(cX, cY);
							}
						}
					}
				}
				else {
					cout << "Unknown BT split" << endl;
				}
			}
			else if (tile.get_split() == Tile::TB_SPLIT) {
				//Tile is split |/|

				if (tile.get_typeW() != TILE_GROUND) {
					for (int cX = startX; cX < endX; cX++) {
						for (int cY = startY; cY < endY; cY++) {
							if (y <= (15 - x)) {
								handleImpassableTerrainTile(cX, cY);
							}
						}
					}
				}
				else if (tile.get_typeE() != TILE_GROUND) {
					for (int cX = startX; cX < endX; cX++) {
						for (int cY = startY; cY < endY; cY++) {
							if (y >= (15 - x)) {
								handleImpassableTerrainTile(cX, cY);
							}
						}
					}
				}
				else {
					cout << "Unknown TB split" << endl;
				}
			}
			else {
				cout << "Unknown tile split" << endl;
			}
		}
	}
}

/*
	This method can be called after creating terrain map to
	clear paths between blocked terrain.
*/
void ChargeMapCreator::postProcessTerrain() {
	for (int cX = 0; cX < c_map->width; cX++) {
		for (int cY = 0; cY < c_map->width; cY++) {
			real8 t_charge = c_map->getTerrainCharge(cX, cY);
			if (t_charge >= TBConfig::params.ChargeMap_PathMinThreshold && t_charge <= TBConfig::params.ChargeMap_PathMaxThreshold) {
				if (clearTile(cX, cY, t_charge)) {
					c_map->setTerrainCharge(cX, cY, 0);
				}
			}
		}
	}
}

/*
	This method can be called after creating terrain map to
	block bad areas in the terrain (i.e. areas surrounded by
	impassable terrain).
*/
void ChargeMapCreator::postProcessTerrain2() {
	real8 preFill = 200;

	for (int cX = 0; cX < c_map->width; cX++) {
		for (int cY = 0; cY < c_map->width; cY++) {
			real8 t_charge = c_map->getTerrainCharge(cX, cY);

			if (t_charge > ChargeMap::CHARGE_IMPASSABLE_TILES && t_charge < -5) {
				int blockCount = 0;
				
				//Passable tile. Check surroundings.

				int distToCheck = 4; //0 for no fill
				//Row 1
				if (isBlocked(cX, cY, -1, -1, distToCheck)) blockCount++;
				if (isBlocked(cX, cY, -0.5, -1, distToCheck)) blockCount++;
				if (isBlocked(cX, cY, 0, -1, distToCheck)) blockCount++;
				if (isBlocked(cX, cY, 0.5, -1, distToCheck)) blockCount++;
				if (isBlocked(cX, cY, 1, -1, distToCheck)) blockCount++;

				//Row 2
				if (isBlocked(cX, cY, -1, -0.5, distToCheck)) blockCount++;
				if (isBlocked(cX, cY, -1, 0, distToCheck)) blockCount++;
				if (isBlocked(cX, cY, -1, 0.5, distToCheck)) blockCount++;

				if (isBlocked(cX, cY, 1, -0.5, distToCheck)) blockCount++;
				if (isBlocked(cX, cY, 1, 0, distToCheck)) blockCount++;
				if (isBlocked(cX, cY, 1, 0.5, distToCheck)) blockCount++;

				//Row 3
				if (isBlocked(cX, cY, -1, 1, distToCheck)) blockCount++;
				if (isBlocked(cX, cY, -0.5, 1, distToCheck)) blockCount++;
				if (isBlocked(cX, cY, 0, 1, distToCheck)) blockCount++;
				if (isBlocked(cX, cY, 0.5, 1, distToCheck)) blockCount++;
				if (isBlocked(cX, cY, 1, 1, distToCheck)) blockCount++;

				//Check if enough directions is blocked => preFill
				if (blockCount >= 8) {
					c_map->setTerrainCharge(cX, cY, preFill);
				}
			}
		}
	}

	//Change preFill values to impassable tile charge
	for (int cX = 0; cX < c_map->width; cX++) {
		for (int cY = 0; cY < c_map->width; cY++) {
			real8 t_charge = c_map->getTerrainCharge(cX, cY);
			if (t_charge == preFill) {
				c_map->setTerrainCharge(cX, cY, ChargeMap::CHARGE_IMPASSABLE_TILES);
			}
		}
	}
}

/*
	Help method for postProcessTerrain2().
	Returns true of the direction [dX,dY] from [cX,cY] has an impassable tile
	within a specified distance (iterations).
*/
bool ChargeMapCreator::isBlocked(int cX, int cY, real8 dX, real8 dY, int iterations) {
	int cnt = 0;
	real8 mX = (real8)cX;
	real8 mY = (real8)cY;

	while (cnt < iterations) {
		real8 t_charge = c_map->getTerrainCharge((int)mX, (int)mY);
		if (t_charge <= ChargeMap::CHARGE_IMPASSABLE_TILES) {
			//Blocked
			return true;
		}

		mX += dX;
		mY += dY;
		cnt++;
	}

	return false;
}

/*
	This is the final postprocess method for terrain map creation.
	It adds an impassable tile charge to the outer bounds of the map.
*/
void ChargeMapCreator::finalPostProcessTerrain() {
	real8 charge = -100;
	for (int cX = 0; cX < c_map->width; cX++) {
		c_map->setTerrainCharge(cX, 0, charge);
		c_map->setTerrainCharge(cX, c_map->height - 1, charge);
	}
	for (int cY = 0; cY < c_map->height; cY++) {
		c_map->setTerrainCharge(0, cY, charge);
		c_map->setTerrainCharge(c_map->width - 1, cY, charge);
	}
}

/*
	Returns true if the terrain tile at the specific point shall be cleared.
	A tile is cleared if it belongs to a path between blocked terrain.
	The charge parameter is the original terrain charge in the specific point.
*/
bool ChargeMapCreator::clearTile(int cX, int cY, real8 charge) {
	int x1 = cX;
	int y1 = cY;
	int x2 = cX;
	int y2 = cY;
	
	x1 = cX - 1;
	y1 = cY;
	x2 = cX + 1;
	y2 = cY;
	if (checkSurroundings(x1, y1, x2, y2, charge)) {
		return true;
	}

	x1 = cX;
	y1 = cY - 1;
	x2 = cX;
	y2 = cY + 1;
	if (checkSurroundings(x1, y1, x2, y2, charge)) {
		return true;
	}

	return false;
}

/*
	Help method for clearTile().
	Returns true if a tile shall be cleared based on how the charges in
	[x1,y1] and [x2,y2] differs from the charge parameter.
*/
bool ChargeMapCreator::checkSurroundings(int x1, int y1, int x2, int y2, real8 charge) {
	real8 ch1 = c_map->getTerrainCharge(x1, y1);
	real8 ch2 = c_map->getTerrainCharge(x2, y2);

	if (ch1 >= TBConfig::params.ChargeMap_PathMinThreshold && ch1 <= TBConfig::params.ChargeMap_PathMaxThreshold) {
		if (ch2 >= TBConfig::params.ChargeMap_PathMinThreshold && ch2 <= TBConfig::params.ChargeMap_PathMaxThreshold) {
			if (ch1 < charge && ch2 < charge) {
				return true;
			}
		}
	}

	return false;
}

/*
	Update the charges for an impassable terrain tile at [cX,cY]
*/
void ChargeMapCreator::handleImpassableTerrainTile(int cX, int cY) {
	int x1 = cX - StaticMap::terrain_w;
	int x2 = cX + StaticMap::terrain_w;
	int y1 = cY - StaticMap::terrain_w;
	int y2 = cY + StaticMap::terrain_w;
	
	for (int mX = x1; mX <= x2; mX++) {
		for (int mY = y1; mY <= y2; mY++) {
			
			if (mX == cX && mY == cY) {
				c_map->updateTerrainCharge(cX, cY, ChargeMap::CHARGE_IMPASSABLE_TILES);
			}
			else {
				real8 charge = (real8)StaticMap::terrain[mX - x1][mY - y1];
				c_map->updateTerrainCharge(mX, mY, charge);
			}
		}
	}
}
