#ifndef ChargeMapCreator_H
#define ChargeMapCreator_H

#include "Game.H"
#include "ChargeMap.H"
#include "ObjMap.H"
#include "GameObj.H"

using namespace std;

/*
	This class is responsible for creating and updating the charge map. It takes the current GameState
	and generates charge areas for all the objects.
	Terrain and bases are generated at startup time and are only updated when necessary (i.e. one or more
	bases destroyed), while the unit charge maps are updated each action frame.
*/
class ChargeMapCreator {

	typedef GameObj* GameObjPtr;
	typedef GameStateModule* GameStateModulePtr;

	public:

		/*
			Constructor.
		*/
		ChargeMapCreator();

		/*
			Destructor.
		*/
		~ChargeMapCreator();

		/*
			Returns false if the charge map needs to be initiated, i.e. the terrain and base
			charges must be generated.
		*/
		bool isInitiated();

		/*
			Returns true if the chargemap is ready for use, i.e. terrain and base charges are generated.
		*/
		bool isReady();

		/*
			Initiates the map. This method is threaded since it usually takes about 150 ms and each action
			frame is 125 ms.
		*/
		void initMap(GameStateModulePtr mState);

		/*
			Init thread run method.
		*/
		static int execute(void *data);

		/*
			Updates the unit charge map to the current game state. Shall be called
			each action frame.
		*/
		void updateUnitMap(const Game& game);

		/*
			Updates the base charge map to the current game state. Shall be called
			when a enemy or friendly base is destroyed.
		*/
		void updateBaseMap(const Game& game);
		
		/*
			Returns a pointer to the charge map.
		*/
		ChargeMap* getMap();

		
		//Global Id for the player
		sint4 playerId;
		//Global Id for the enemy
		sint4 enemyId;

		/*
			Updates own-, enemy- and neutral units in the ObjMap.
		*/
		void updateObjMap(const Game& game);

		/*
			Updates own- and enemy bases in the ObjMap.
		*/
		void updateObjBaseMap(const Game& game);

		/*
			Returns the ObjMap.
		*/
		ObjMap* getObjMap();

		//The ObjMap.
		ObjMap* obj_map;

	protected:
		
		/*
			Creates charge map for the terrain.
		*/
		void createTerrainMap(const Game& game);

		/*
			This method can be called after creating terrain map to
			clear paths between blocked terrain.
		*/
		void postProcessTerrain();

		/*
			This method can be called after creating terrain map to
			block bad areas in the terrain (i.e. areas surrounded by
			impassable terrain).
		*/
		void postProcessTerrain2();

		/*
			This is the final postprocess method for terrain map creation.
			It adds an impassable tile charge to the outer bounds of the map.
		*/
		void finalPostProcessTerrain();

		/*
			Help method for postProcessTerrain2().
			Returns true of the direction [dX,dY] from [cX,cY] has an impassable tile
			within a specified distance (iterations).
		*/
		bool isBlocked(int cX, int cY, real8 dX, real8 dY, int iterations);

		/*
			Returns true if the terrain tile at the specific point shall be cleared.
			A tile is cleared if it belongs to a path between blocked terrain.
			The charge parameter is the original terrain charge in the specific point.
		*/
		bool clearTile(int cX, int cY, real8 charge);

		/*
			Help method for clearTile().
			Returns true if a tile shall be cleared based on how the charges in
			[x1,y1] and [x2,y2] differs from the charge parameter.
		*/
		bool checkSurroundings(int x1, int y1, int x2, int y2, real8 charge);

		//Init map thread
		SDL_Thread *my_thread;
		
		//The charge map
		ChargeMap* c_map;

		//Current game state. Only used in the init thread.
		GameStateModulePtr state;

		//Set to true if the client is started in debug mode
		bool debug;

		/*
			Update the charges for an impassable terrain tile at [cX,cY]
		*/
		void handleImpassableTerrainTile(int cX, int cY);
		

};

#endif
