#include "ClientGUI.H"
#include "SDL_gfxPrimitives.h"
#include "Options.H"

using namespace std;

/*
	Constructor.
*/
ClientGUI::ClientGUI() {
	bool debug;
	Options::get("-debug", debug);
	
	if (debug) {
		screen = 0;
		gfx_mag_frac = 16;
	
		s_x = 769;
		s_y = 769;
		adj_x = s_x/32;
		adj_y = s_y/32;
	
		// open screen
		
		// Have a preference for 8-bit, but accept any depth
	
		screen = SDL_SetVideoMode(s_x, s_y, 0, SDL_HWSURFACE | SDL_ANYFORMAT);
		if (screen == 0) {
			//   cerr << "Couldn't set " << s_x << "x" << s_y << "x8 video mode: " << SDL_GetError() << endl;
			exit(1);
		}
	
		ostringstream oss;
		oss << "Displaying Potential Fields";
	
		SDL_WM_SetCaption(oss.str().c_str(), 0);
	}
}

/*
	Updates the ChargeMap.
*/
void ClientGUI::setRef(ChargeMap* map, ObjMap* objmap, GameStateModule* state, FireCoordinator* coord) {
	c_map = map;
	obj_map = objmap;
	gamestate = state;
	coordinator = coord;
}

/*
	Updates the GUI.
*/
void ClientGUI::display() {
	if (SDL_MUSTLOCK(screen) && SDL_LockSurface(screen) < 0 ) {
		fprintf(stderr, "Can't lock screen: %s\n", SDL_GetError());
		return;
	}
	
	//Update GFX
	for (int cX = 0; cX < c_map->width; cX++) {
		for (int cY = 0; cY < c_map->height; cY++) {
			int mX = cX * 8 + 4;
			//if (mX > 1017) mX = 1017;
			int mY = cY * 8 + 4;
			//if (mY > 1017) mY = 1017;
			real8 charge = c_map->getP(mX, mY, true, 0, gamestate->get_game(), obj_map);
			
			int w = 6;
			SDL_Rect rect;
			rect.x = cX * w;
			rect.y = cY * w;
			rect.w = w;
			rect.h = w;

			SDL_FillRect(screen, &rect, getColor(charge));
		}
	}

	vector<ObjFuturePos>* units = obj_map->getUnitsList();
	
	//Draw attack lines
	vector<Attack>::iterator dit = attacks.begin();
	while (dit != attacks.end()) {
		if (dit->isActive()) {
			vector<ObjFuturePos>::iterator it2 = units->begin();
			while (it2 != units->end()) {
				if(it2->isThisObj(dit->unitId)) {
					drawAttackLine(it2->x, it2->y, dit->targetId, dit->cnt);
				}
				it2++;
			}
			dit->fade();
		}
		dit++;
	}
	//

	//Own Units
	RGB ergb(0, 255, 0);
	vector<ObjFuturePos>::iterator uit = units->begin();
	while (uit != units->end()) {
		//Unit
		filledCircleColor(screen, toScreenCoord(uit->x), toScreenCoord(uit->y), (sint2)2, ergb.color_alpha());

		//Attacks
		int targetId = coordinator->getTargetId(uit->objId);
		if (targetId > 0) {
			//drawAttackLine(uit->x, uit->y, targetId);
			updateAttack(uit->objId, targetId);
		}
		uit++;
	}

	//Enemy Units
	RGB rgb(255, 0, 0);
	vector<ObjPos>* eunits = obj_map->getEnemyUnitsList();
	vector<ObjPos>::iterator euit = eunits->begin();
	while (euit != eunits->end()) {
		filledCircleColor(screen, toScreenCoord(euit->x), toScreenCoord(euit->y), (sint2)2, rgb.color_alpha());
		euit++;
	}

	if (SDL_MUSTLOCK(screen)) SDL_UnlockSurface(screen);

	SDL_Flip(screen);
	SDL_UpdateRect(screen, 0, 0, 0, 0);
}

void ClientGUI::updateAttack(int unitId, int targetId) {
	bool found = false;
	vector<Attack>::iterator it = attacks.begin();
	while (it != attacks.end()) {
		if (it->isThisObj(unitId)) {
			it->targetId = targetId;
			it->reset();
			found = true;
		}
		it++;
	}

	//Unit struct not found. Add new
	if (!found) {
		Attack atk;
		atk.unitId = unitId;
		atk.targetId = targetId;
		atk.reset();
		attacks.push_back(atk);
	}
}

void ClientGUI::drawAttackLine(int x, int y, int targetId, int fadeVal) {
	int v = 255 * fadeVal / 10;
 	RGB rgb(v, v, v);
	vector<ObjPos>* eunits = obj_map->getEnemyUnitsList();
	vector<ObjPos>::iterator euit = eunits->begin();
	while (euit != eunits->end()) {
		if (euit->isThisObj((uint4)targetId)) {
			lineColor(screen, toScreenCoord(x), toScreenCoord(y), toScreenCoord(euit->x), toScreenCoord(euit->y), rgb.color_alpha());
			return;
		}
		euit++;
	}


	vector<BasePos>* ebases = obj_map->getEnemyBasesList();
	vector<BasePos>::iterator ebit = ebases->begin();
	while (ebit != ebases->end()) {
		if (ebit->isThisObj((uint4)targetId)) {
			lineColor(screen, toScreenCoord(x), toScreenCoord(y), toScreenCoord(ebit->x), toScreenCoord(ebit->y), rgb.color_alpha());
			return;
		}
		ebit++;
	}
}

int ClientGUI::toScreenCoord(int coord) {
	double d = round((double)coord * (double)s_x / (double)1024);
	return (int)d - 0; //d - 6
}

/*
	Returns a display color for the specified charge.
*/
Uint32 ClientGUI::getColor(real8 charge) {
	//Update
	if (charge <= ChargeMap::CHARGE_IMPASSABLE_TILES / 4) {
		//Impassable terrain
		RGB rgb(44, 44, 44);
		return rgb.color();
	}
	if (charge == 0) {
		RGB rgb(0, 0, 0);
		return rgb.color();
	}
	
	if (charge < 0) {
		int R = (int)((float)(charge * -1) * 6);
		int G = (int)(R / 4);
		int B = (int)(R / 4);

		if (R < 1) R = 1;
		if (R > 254) R = 254;
		if (G < 1) G = 1;
		if (G > 254) G = 254;
		if (B < 1) B = 1;
		if (B > 254) B = 254;

		RGB rgb(R, G, B);
		return rgb.color();
	}
	if (charge > 0) {
		//int B = (int)((float)charge * 4);
		int B = (int)((float)charge * charge / 5);
		int R = (int)(B / 6); // /4
		int G = (int)(B / 6);

		if (R < 1) R = 1;
		if (R > 254) R = 254;
		if (G < 1) G = 1;
		if (G > 254) G = 254;
		if (B < 1) B = 1;
		if (B > 254) B = 254;
		
		RGB rgb(R, G, B);
		return rgb.color();
	}
	
	return 0;
}

