#include "ObjMap.H"
#include "ChargeMap.H"
#include "EventTimer.H"
#include "TBConfig.H"
#include "Blackboard.H"

ObjMap::ObjMap() {
	
}

int ObjMap::getOU() {
	return units.size();
}

int ObjMap::getEU() {
	return enemies.size();
}

int ObjMap::getOB() {
	return bases.size();
}

int ObjMap::getEB() {
	return enemyBases.size();
}

vector<ObjFuturePos>* ObjMap::getUnitsList() {
	return &units;
}

vector<ObjPos>* ObjMap::getEnemyUnitsList() {
	return &enemies;
}

vector<BasePos>* ObjMap::getEnemyBasesList() {
	return &enemyBases;
}

/*
	Sets all own and enemy units as destroyed. Shall be called before updating the unit positions
	each frame. A unit which has its position updated are set as not destroyed, and after the
	position updates all units which still has the destroyed flag set can be removed.
*/
void ObjMap::setAllDestroyed() {
	vector<ObjPos>::iterator it = enemies.begin();
	while (it != enemies.end()) {
		(*it).destroyed = true;
		it++;
	}

	vector<ObjFuturePos>::iterator uit = units.begin();
	while (uit != units.end()) {
		(*uit).destroyed = true;
		uit++;
	}
}

/*
	Removes destroyed own and enemy units.
	Must be used with the setAllDestroyed() method.
*/
void ObjMap::cleanUp() {
	for (int i = 0; i < (int)enemies.size(); i++) {
		if (enemies[i].destroyed) {
			vector<ObjPos>::iterator it = enemies.begin();
			it += i;
			enemies.erase(it);
			i--;
		}
	}

	for (int i = 0; i < (int)units.size(); i++) {
		if (units[i].destroyed) {
			vector<ObjFuturePos>::iterator it = units.begin();
			it += i;
			units.erase(it);
			i--;
		}
	}
}

/*
	Updates the position for sheep [objId].
*/
void ObjMap::updateSheepPosition(uint4 game_tick, uint4 objId, int x, int y) {
	//Check if struct for this unit exists and, if found, update
	bool found = false;
	vector<ObjPos>::iterator it = sheep.begin();
	while (it != sheep.end()) {
		if ((*it).isThisObj(objId)) {
			(*it).update(game_tick, x, y);
			found = true;
		}
		it++;
	}

	//Unit struct not found. Add new
	if (!found) {
		ObjPos pos;
		pos.objId = objId;
		pos.x = x;
		pos.y = y;
		pos.gametick = game_tick;
		sheep.push_back(pos);
	}
}

/*
	Returns the potential generated by sheep in the point (pX,pY).
*/
real8 ObjMap::getSheepCharge(int pX, int pY, uint4 game_tick) {
	real8 p = 0;

	vector<ObjPos>::iterator it = sheep.begin();
	while (it != sheep.end()) {
		if (ChargeMap::MHdistance(pX, pY, (*it).x, (*it).y) <= 26) {
			p += calcSheepCharge(pX, pY, (*it).x, (*it).y);
		}

		it++;
	}
	
	return p;
}

/*
	Returns the potential a sheep at (objX,objY) generates in the point (x,y).
*/
real8 ObjMap::calcSheepCharge(int x, int y, int objX, int objY) {
	real8 p = 0;
	real8 dist = ChargeMap::distance(x, y, objX, objY);

	if (bases.size() == 0) {
		//Game 4 -> Smaller sheep field
		if (dist <= 10) {
			if (dist <= 4) { //4  radius(sheep) = 4
				p = - 20;
			}
			else {
				p = -(20 - 2 * dist);
				if (p > 0) p = 0;
			}
		}
	}
	else {
		if (dist <= 8) {
			p = -20;
		}
		else {
			p = (real8)2 * dist - 26;
			if (p > 0) p = 0;
		}
		p = p * TBConfig::params.Sheep_w;
	}
	
	return p;
}

/*
	Updates the position for enemy unit [objId].
*/
void ObjMap::updateEnemyUnitPosition(uint4 game_tick, uint4 objId, sint4 objType, int x, int y) {
	//Check if struct for this unit exists and, if found, update
	bool found = false;
	vector<ObjPos>::iterator it = enemies.begin();
	while (it != enemies.end()) {
		if ((*it).isThisObj(objId)) {
			(*it).update(game_tick, x, y);
			found = true;
			return;
		}
		it++;
	}

	//Unit struct not found. Add new
	if (!found) {
		ObjPos pos;
		pos.objId = objId;
		pos.type = objType;
		pos.x = x;
		pos.y = y;
		pos.gametick = game_tick;
		enemies.push_back(pos);
	}
}

/*
	Returns the potential generated by enemy units in the point (pX,pY).
*/
real8 ObjMap::getEnemyUnitCharge(int pX, int pY, bool offensive, uint4 game_tick) {
	real8 p = 0;

	vector<ObjPos>::iterator it = enemies.begin();
	while (it != enemies.end()) {
		if (!(*it).destroyed && (*it).isMarine()) {
			if (offensive) {
				real8 mp = calcOffensiveMarineCharge(pX, pY, (*it).x, (*it).y);
				if (TBConfig::hasAddition1()) {
					if (p < mp) {
						p = mp;
					}
				}
				else {
					p += mp;
				}
			}
			if (!offensive) {
				p += calcDefensiveMarineCharge(pX, pY, (*it).x, (*it).y);
			}
		}

		if (!(*it).destroyed && (*it).isTank()) {
			if (offensive) {
				real8 mp = calcOffensiveTankCharge(pX, pY, (*it).x, (*it).y);
				if (TBConfig::hasAddition1()) {
					if (p < mp) {
						p = mp;
					}
				}
				else {
					p += mp;
				}
			}
			if (!offensive) {
				p += calcDefensiveTankCharge(pX, pY, (*it).x, (*it).y);
			}
		}

		it++;
	}
	
	return p;
}

/*
	Returns the potential a marine at (objX,objY) generates in the point (x,y).
*/
real8 ObjMap::calcOffensiveMarineCharge(int x, int y, int objX, int objY) {
	real8 p = 0;
	real8 dist = ChargeMap::distance(x, y, objX, objY);
	if (dist <= 63) {
		p = dist * 0.8 * 5;
	}
	else if (dist > 63 && dist <= 64) {
		p = 65 * 5;
	}
	else {
		p = (50 - (dist - 64) * 0.05) * 5;
		if (p < 0) p = 0;
	}

	return p;
}

/*
	Returns the potential a marine at (objX,objY) generates in the point (x,y).
*/
real8 ObjMap::calcDefensiveMarineCharge(int x, int y, int objX, int objY) {
	if (ChargeMap::MHdistance(x, y, objX, objY) > 144) return 0;

	real8 p = 0;
	real8 dist = ChargeMap::distance(x, y, objX, objY);
	if (dist <= 72) { //72
		p = (-60 + (dist * 0.8)) * 10;
	}
	
	return p;
}

/*
	Returns the potential a tank at (objX,objY) generates in the point (x,y).
*/
real8 ObjMap::calcOffensiveTankCharge(int x, int y, int objX, int objY) {
	real8 p = 0;
	real8 dist = ChargeMap::distance(x, y, objX, objY);

	if (dist <= 110) {
		p = (real8)240 / (real8)110 * dist;
	}
	else if (dist > 110 && dist <= 112) {
		p = 240;
	}
	else {
		p = 240 - (dist - 112) * (real8)0.24;
		if (p < 0) p = 0;
	}
	p = p * TBConfig::params.OffTank_w;
	
	return p;
}

/*
	Returns the potential a tank at (objX,objY) generates in the point (x,y).
*/
real8 ObjMap::calcDefensiveTankCharge(int x, int y, int objX, int objY) {
	real8 dist = ChargeMap::distance(x, y, objX, objY);
	real8 p = -800 + dist * 6.4;
	if (p > 0) {
		p = 0;
	}
	p = p * TBConfig::params.DefTank_w;
	
	return p;
}

/*
	Updates the position for own unit [objId].
*/
void ObjMap::updateUnitPosition(uint4 game_tick, uint4 objId, sint4 objType, int x, int y) {
	//Check if struct for this unit exists and, if found, update
	bool found = false;
	vector<ObjFuturePos>::iterator it = units.begin();
	while (it != units.end()) {
		if ((*it).isThisObj(objId)) {
			(*it).update(game_tick, x, y);
			found = true;

			//Update trail
			(*it).trail.update(x, y, game_tick);
			
			return;
		}
		it++;
	}

	//Unit struct not found. Add new
	if (!found) {
		ObjFuturePos pos;
		pos.objId = objId;
		pos.type = objType;
		pos.x = x;
		pos.y = y;
		pos.gametick = game_tick;
		units.push_back(pos);

		//Init trail
		Trail trail;
		trail.init(x, y, game_tick);
		pos.trail = trail;
	}
}

/*
	Sets the position the unit [objId] is moving to.
*/
void ObjMap::setNextUnitPosition(uint4 objId, int nextX, int nextY) {
	vector<ObjFuturePos>::iterator it = units.begin();
	while (it != units.end()) {
		if ((*it).isThisObj(objId)) {
			(*it).nextX = nextX;
			(*it).nextY = nextY;
			return;
		}
		it++;
	}
}

/*
	Returns the potential generated by own units in the point (pX,pY).
*/
real8 ObjMap::getUnitCharge(int pX, int pY, uint4 objId, uint4 game_tick) {
	real8 p = 0;

	vector<ObjFuturePos>::iterator it = units.begin();
	while (it != units.end()) {
		if (!(*it).destroyed && (*it).isMarine() && (*it).objId != objId) {
			p += calcMarineUnitCharge(pX, pY, (*it).x, (*it).y);
			if ((*it).hasNextPos()) {
				p += calcMarineUnitCharge(pX, pY, (*it).nextX, (*it).nextY);
			}
		}
		if (!(*it).destroyed && (*it).isTank() && (*it).objId != objId) {
			p += calcTankUnitCharge(pX, pY, (*it).x, (*it).y);
			if ((*it).hasNextPos()) {
				p += calcTankUnitCharge(pX, pY, (*it).nextX, (*it).nextY);
			}
		}

		if (TBConfig::hasAddition4()) {
			//Trail charge
			if ((*it).isThisObj(objId)) {
				real8 v = it->trail.getP(pX, pY, game_tick);
				p += it->trail.getP(pX, pY, game_tick);
			}
		}

		it++;
	}
	
	return p;
}

/*
	Returns the potential an own marine at (objX,objY) generates in the point (x,y).
*/
real8 ObjMap::calcMarineUnitCharge(int x, int y, int objX, int objY) {
	if (ChargeMap::MHdistance(x, y, objX, objY) > 20) return 0;

	real8 p = 0;
	real8 dist = ChargeMap::distance(x, y, objX, objY);
	if (dist <= 8) { //radius(marine) = 4
		p = p - 20;
	}
	else {
		real8 mp = 20 - 2 * dist; //20 - 2 * dist
		if (mp < 0) mp = 0;

		p = p - mp;
	}
	
	return p;
}

/*
	Returns the potential generated by the marine coordination field in the point (pX,pY).
*/
real8 ObjMap::getMarineCoordCharge(int pX, int pY, uint4 game_tick, int map_width) {
	real8 p = 0;

	int diffX = pX - 1;
	bool startAtLeft = false;
	if (pX <= map_width / 2) {
		startAtLeft = true;
	}
	
	if (!startAtLeft) {
		diffX = map_width - pX - 1;
	}
	
	if (diffX <= 224) {
		p += (real8)diffX / (real8)8 * (real8)3;
	}
	else if (diffX > 224 && diffX <= 248) {
		p += 92;
	}
	else if (diffX > 248 && diffX < 272) {
		p += 92 - (diffX - 248) * 0.3;
	}
	
	return p;
}

/*
	Returns the potential an own tank at (objX,objY) generates in the point (x,y).
*/
real8 ObjMap::calcTankUnitCharge(int x, int y, int objX, int objY) {
	real8 p = 0;
	real8 dist = ChargeMap::distance(x, y, objX, objY);

	if (dist <= 14) {
		p = p - 20;
	}
	else  {
		real8 mp = 32 - dist * 2;
		if (mp < 0) mp = 0;
		p = p - mp;
	}
	p = p * TBConfig::params.Tank_w;
	
	return p;
}

/*
	Removes all elements in the own- and enemy bases vectors.
*/
void ObjMap::clearBaseLists() {
	bases.clear();
	enemyBases.clear();
}

/*
	Adds the base [objId] to a bases vector.
*/
void ObjMap::addBase(uint4 objId, int x, int y, int x1, int y1, int x2, int y2, bool own) {
	BasePos pos;
	pos.objId = objId;
	pos.x = x;
	pos.y = y;
	pos.x1 = x1;
	pos.y1 = y1;
	pos.x2 = x2;
	pos.y2 = y2;
	if (own) {
		bases.push_back(pos);
	}
	else {
		enemyBases.push_back(pos);
	}
}

/*
	Returns the potential generated by own bases in the point (pX,pY).
*/
real8 ObjMap::getBaseCharge(int pX, int pY) {
	real8 p = 0;

	vector<BasePos>::iterator it = bases.begin();
	while (it != bases.end()) {
		p += calcBaseCharge(pX, pY, (*it).x, (*it).y, (*it).isPositionInside(pX, pY));
		it++;
	}

	return p;
}

/*
	Returns the potential an own base at (objX,objY) generates in the point (x,y).
*/
real8 ObjMap::calcBaseCharge(int x, int y, int objX, int objY, bool inside) {
	//Manually optimized version
	/*real8 p = 0;
	if (inside) {
		p = ChargeMap::CHARGE_IMPASSABLE_TILES;
	}
	else {
		real8 dist = ChargeMap::distance(x, y, objX, objY);

		//Base width = 62
		p = 6 * dist - 258;
		if (p > 0) p = 0;
	}
	return p;*/

	//GA optimized version
	real8 p = 0;
	if (inside) {
		p = ChargeMap::CHARGE_IMPASSABLE_TILES * 10;
	}
	else {
		real8 dist = ChargeMap::distance(x, y, objX, objY);

		//Base width = 62
		p = (real8)(6 * dist - 258);
		if (p > 0) p = 0;
	}
	p = p * TBConfig::params.Base_w;

	return p;
}

/*
	Returns the potential generated by enemy bases in the point (pX,pY).
*/
real8 ObjMap::getEnemyBaseCharge(int pX, int pY) {
	real8 p = 0;

	vector<BasePos>::iterator it = enemyBases.begin();
	while (it != enemyBases.end()) {
		if (Blackboard::state == Blackboard::STATE_HUNT_UNITS) {
			p += calcBaseCharge(pX, pY, (*it).x, (*it).y, (*it).isPositionInside(pX, pY));
		}
		else {
			real8 mp = calcEnemyBaseCharge(pX, pY, (*it).x, (*it).y, (*it).isPositionInside(pX, pY));
			if (TBConfig::hasAddition1()) {
				if (mp > p) {
					p = mp;
				}
			}
			else {
				p += mp;
			}
		}
		it++;
	}

	return p;
}

/*
	Returns the potential an enemy base at (objX,objY) generates in the point (x,y).
*/
real8 ObjMap::calcEnemyBaseCharge(int x, int y, int objX, int objY, bool inside) {
	real8 p = 0;
	if (inside) {
		p = 0;
	}
	else {
		real8 dist = ChargeMap::distance(x, y, objX, objY) - (real8)31; //Base width = 64
		
		if (dist <= 110) {
			p = (real8)360 / (real8)110 * dist;
		}
		else if (dist <= 112) {
			p = 360;
		}
		else {
			p = (real8)240 - (dist - 112) * 0.32;
			if (p < 0) p = 0;
		}

		p = p * TBConfig::params.OffBase_w;
	}
	return p;
}

/*
	Predicts the next position (x or y) a unit might move to.
*/
int ObjMap::predictNext(uint4 game_tick, int p_minusOne, int p_minusTwo, int gametick_minusOne, int gametick_minusTwo) {
	//Have not found two positions yet. Cannot predict.
	if (p_minusOne == -1 || p_minusTwo == -1) {
		return -1;
	}

	real8 diff = p_minusOne - p_minusTwo;

	real8 upd_interval = gametick_minusOne - gametick_minusTwo;
	real8 cur_interval = game_tick - gametick_minusOne;
	real8 mul = cur_interval / upd_interval;
	if (mul > 1) {
		//Interval is larger than prev. Set to 1
		mul = 1;
	}

	real8 actual_diff = diff * mul;

	int res = p_minusOne + (int)round(actual_diff);

	return res;
}
