
#include "TBEventHandler.H"
#include "GameStateModule.H"
#include "GfxModule.H"
#include "Game.H"
#include "GameChanges.H"
#include "GameTile.H"
#include "ServerObjData.H"
#include "Options.H"
#include "UnitAgent.H"
#include "Blackboard.H"
#include "EventTimer.H"

using namespace std;

bool TBEventHandler::handle_event(const Event& e) {
	if (e.get_who() == GameStateModule::FROM) {
    
		if (e.get_what() == GameStateModule::STOP_MSG) {
			cout << "Game stopped" << endl;
			exit(0);
		}
    
		if (e.get_what() == GameStateModule::READ_ERROR_MSG) {
			cout << "Error occured: " << state.error << endl;
			exit(1);
		}
    
		if (e.get_what() == GameStateModule::VIEW_MSG) {
      
			compute_actions();
			// Send my actions
			state.gsm->send_actions();
			
			Game &game = state.gsm->get_game();	

			// gfx and frame statistics
      
			if (state.gfxm) state.gfxm->process_changes();
			
			sint4 vf = game.get_view_frame();
			sint4 af = game.get_action_frame();

			if (state.gfxm) {
	
				// don't draw if we are far behind
				if (abs(vf-af) < 10) {
					state.gfxm->draw();
					state.just_drew = true;
				}
			}

			return true;
		}
	}
	
	return false;
}


void TBEventHandler::compute_actions() {
	//EventTimer::start();
	const GameChanges &changes = state.gsm->get_changes();
	
	bool baseDestroyed = false;
	bool unitDestroyed = false;

	if (!changes.dead_objs.empty()) {
		FORALL (state.gsm->get_changes().dead_objs, it) {
        	GameObj *gob = (*it)->get_GameObj();
			std::string objType = gob->bp_name();
			//cout << "Dead object: " << objType << endl;
			if (objType == "controlCenter") {
				//A base has been destroyed
				baseDestroyed = true;
			}
			if (objType == "tank" || objType == "marine") {
				//A unit has been destroyed
				unitDestroyed = true;
			}
		}
	}

	//Current game state
	const Game &game = state.gsm->get_game();
	
	//This is the unique id for this player
	const sint4 cid = game.get_client_player();

	const Game::ObjCont &objs = game.get_objs(cid);
	
	bool debug;
	Options::get("-debug", debug);
	
	if (!mapCreator.isInitiated()) {
		//We need to init the mapCreator
		mapCreator.initMap(state.gsm);
		gui.setRef(mapCreator.getMap(), mapCreator.getObjMap(), state.gsm, &coordinator);
	}

	
	if (mapCreator.isReady()) {
		//Update game tactics
		if (!Blackboard::stateLocked()) {
			Blackboard::updateState(game);
		}

		//If a base has been destroyed, we need to update the base map
		if (baseDestroyed) {
			mapCreator.updateBaseMap(game);
		}
		
		//Everything is ready. Update the unit map each frame
		mapCreator.updateUnitMap(game);
	}

	//Iterate through this players' objects
	if (mapCreator.isReady()) {
		//Coordinate fire at enemy targets
		//T4
		coordinator.coordinate_fire(state.gsm, &mapCreator);
		//T4 totalCPU=21 avgCPU=0.04 calls=510
		
		FORALL (objs, it) {
			GameObj *gob = (*it)->get_GameObj();
			if (gob && gob->sod.in_game && *gob->sod.x >= 0 && !gob->is_dead()) {
				std::string objType = gob->bp_name();
				
				//This GameObj is in the game
				if (objType == "tank") {
					UnitAgent::compute_actions(gob, state.gsm, &mapCreator, &coordinator, &rand);
				}
				if (objType == "marine") {
					//T1
					UnitAgent::compute_actions(gob, state.gsm, &mapCreator, &coordinator, &rand);
					//T1 totalCPU=430 avgCPU=0.019 calls=22692
				}
			}
		}
	}
	
	sint4 af = game.get_action_frame();

	bool writeStatus = false;
	if (af % 100 == 0) writeStatus = true;
	if (baseDestroyed) writeStatus = true;
	if (unitDestroyed) writeStatus = true;
	
	//This writes unit status necessary for the GA optimization / tournament tool.
	/*if (mapCreator.isReady() && writeStatus) {
		cout << "" << endl;
		cout << "AF=" << af << " playerId=" << mapCreator.playerId << endl;
	}*/
	
	if (mapCreator.isReady() && unitDestroyed) {
		mapCreator.getObjMap()->cleanUp();
	}

	//EventTimer::stop(0);

	//if (af % 10 == 0) {
		//dumpInfo(af);
	//}

	if (debug) {
		gui.display();
	}
}

void TBEventHandler::dumpInfo(int af) {
	ofstream fout;
	fout.open("html/tb3_data.txt", ios::app);

	sint4 seed;
  	Options::get("-seed", seed);

	fout << "seed=" << seed << " ";
	fout << "af=" << af << " ";
	fout << "ou=" << mapCreator.getObjMap()->getOU() << " ";
	fout << "ob=" << mapCreator.getObjMap()->getOB() << " ";
	fout << "eu=" << mapCreator.getObjMap()->getEU() << " ";
	fout << "eb=" << mapCreator.getObjMap()->getEB() << " ";
	fout << "avgTime=" << EventTimer::getAvgTime() << "\n";
	fout.close();
}
