#include "UnitAgent.H"
#include "StaticMap.H"
#include "Blackboard.H"
#include "TBConfig.H"
#include "FireCoordinator.H"
#include "Options.H"

const int UnitAgent::TANK_FIRERANGE = 112;
const int UnitAgent::TANK_HP = 150;
const int UnitAgent::TANK_ARMOR = 1;
const sint4 UnitAgent::TANK_MIN_DMG = 26;
const sint4 UnitAgent::TANK_MAX_DMG = 34;

const int UnitAgent::MARINE_FIRERANGE = 64;
const int UnitAgent::MARINE_HP = 80;
const int UnitAgent::MARINE_ARMOR = 1;
const sint4 UnitAgent::MARINE_MIN_DMG = 5;
const sint4 UnitAgent::MARINE_MAX_DMG = 7;

const int UnitAgent::CONTROLCENTER_HP = 1200;
const int UnitAgent::CONTROLCENTER_ARMOR = 0;

/*
	Compute actions for this tank object.
*/
void UnitAgent::compute_actions(GameObjPtr unit, GameStateModulePtr state, ChargeMapCreatorPtr mapCreator, FireCoordinatorPtr coordinator, Random* rand) {
	Game& game = state->get_game();
	
	uint4 objId = game.get_cplayer_info().get_id(unit);
	int unitX = *unit->sod.x;
	int unitY = *unit->sod.y;
	
	//Get which mode (offensive/defensive) to use
	bool offensive = getMode(unit, state);
	
	//Fire if any targets are within range
	bool fired = canFire(unit, unitX, unitY, objId, state, offensive, coordinator);
	if (fired) {
		return;
	}
	
	//No targets to fire at, check if we can move
	bool moved = canMove(unit, unitX, unitY, objId, offensive, mapCreator, game);
	if (moved) return;
}

/*
	Checks if the unit can move to the specified position.
*/
bool UnitAgent::canDoMove(uint4 objId, int fromX, int fromY, int toX, int toY, ChargeMapCreatorPtr mapCreator, const Game& game) {
	//Check boundaries
	if (toX < 4) return false;
	if (toX > mapCreator->getMap()->getRealWidth() - 5) return false;
	if (toY < 4) return false;
	if (toY > mapCreator->getMap()->getRealHeight() - 5) return false;

	int diffX = toX - fromX;
	int diffY = toY - fromY;
	int divider = max(abs(diffX), abs(diffY));

	int cnt = 1; //Don't check the units current position (it is passable since the unit is there...)
	int cX = fromX;
	int cY = fromY;

	real8 limit = (real8)ChargeMap::CHARGE_IMPASSABLE_TILES * 0.5;
	
	while (cnt <= divider) {
		int mX = cX + calcNext(cnt, diffX, divider);
		int mY = cY + calcNext(cnt, diffY, divider);

		real8 charge = mapCreator->getMap()->getTerrainCharge(mX, mY);
		if (charge <= limit) {
			//Impassable tile found
			
			//Special case: If units are at the edges of the map, it can always move away from the edge.
			//This avoids units getting stuck at edges
			if (boundaryMove(fromX, fromY, toX, toY, mapCreator->getMap())) {
				return true;
			}

			return false;
		}
		
		cnt++;
	}

	return true;
}

bool UnitAgent::boundaryMove(int fromX, int fromY, int toX, int toY, ChargeMapPtr chargeMap) {
	int d = 10;

	if (fromY == toY) {
		//Check vertical boundary move
		if (toX > fromX && toX <= d) {
			return true;
		}
		if (toX < fromX && toX >= chargeMap->getRealWidth() - d) {
			return true;
		}
	}

	if (fromX == toX) {
		//Check horisontal boundary move
		if (toY > fromY && toY <= d) {
			return true;
		}
		if (toY < fromY && toY >= chargeMap->getRealHeight() - d) {
			return true;
		}
	}

	return false;
}

int UnitAgent::calcNext(int cnt, int diffVal, int divider) {
	real8 next = (real8) cnt * (real8)diffVal / (real8)divider;

	int val = (int)(next * 10);
	int res = (int)next;

	if (val % 10 >= 5) res++;

	return res;
}

/*
	Checks if there are better positions nearby to move to.
	Returns true if unit moves, false if not.
*/
bool UnitAgent::canMove(GameObjPtr unit, int unitX, int unitY, uint4 objId, bool offensive, ChargeMapCreatorPtr mapCreator, const Game& game) {
	//If no targets are within range, check if there are better positions to move to
	ChargeMapPtr chargeMap = mapCreator->getMap();

	int maxSpeed = *unit->sod.max_speed;
	if (Blackboard::getType(unit) == Blackboard::TYPE_MARINE) {
		//If marines move at slow speed they tend to group together more effectively
		maxSpeed = 1;
	}
	
	//Calculate which tiles to check
	int x1 = unitX - maxSpeed;
	int x2 = unitX + maxSpeed;
	int y1 = unitY - maxSpeed;
	int y2 = unitY + maxSpeed;

	//Iterate through the tiles to get the highest (=best) charge
	//Start with the charge for current tile
	real8 bestCharge = chargeMap->getP(unitX, unitY, offensive, objId, game, mapCreator->getObjMap());
	
	int bestX = unitX;
	int bestY = unitY;
	//real8 startP = bestCharge;
	
	//Get the surrounding tile with highest charge
	for (int cX = x1; cX <= x2; cX++) {
		for (int cY = y1; cY <= y2; cY++) {
			if (cX != unitX || cY != unitY) {
				if (canDoMove(objId, unitX, unitY, cX, cY, mapCreator, game)) {
					real8 cCharge = chargeMap->getP(cX, cY, offensive, objId, game, mapCreator->getObjMap());
					
					if (cCharge > bestCharge) {
						bestCharge = cCharge;
						bestX = cX;
						bestY = cY;
					}
				}
			}
		}
	}

	//If we've found a surrounding tile with higher charge than current tile, move there
	if (bestX != unitX || bestY != unitY) {
		int useSpeed = *unit->sod.max_speed;
		if (chargeMap->distance(unitX, unitY, bestX, bestY) < useSpeed) {
			useSpeed = chargeMap->distance(unitX, unitY, bestX, bestY);
		}
		mapCreator->getObjMap()->setNextUnitPosition(objId, bestX, bestY);

		Vector<sint4> params;
		params.push_back(bestX);
		params.push_back(bestY);
		params.push_back(useSpeed);

		unit->set_action("move", params);

		//Unit has moved
		return true;
	}

	//Unit has not moved
	return false;
}

/*
	Decides which mode (offensive = true, defensive = false) to use for a tank.
*/
bool UnitAgent::getMode(GameObjPtr unit, GameStateModulePtr state) {
	bool offensive = true;

	if (TBConfig::hasAddition3()) {
		Game& game = state->get_game();
		//Addition #4: Units are always in defensive move (tries to stay out of reach for fire) if in cooldown
		sint4 lastFired = unit->component("weapon")->get_int("last_fired");
		if (lastFired >= 0) {
			uint4 game_tick = game.tick();
			sint4 cooldown = unit->component("weapon")->get_int("cooldown");
			int timeToShoot = cooldown - (game_tick - (uint4)lastFired);
			if (timeToShoot > 0) {
				offensive = false;
				return offensive;
			}
		}
	}

	return offensive;
}

/*
	Checks for enemy targets within range and fires if any targets found.
	Returns true if fire at a target.

	This method uses a global target coordination to decide what target to attack (if more than one is within firerange).
*/
bool UnitAgent::canFire(GameObjPtr unit, int unitX, int unitY, uint4 objId, GameStateModulePtr state, bool offensive, FireCoordinatorPtr coordinator) {
	Game& game = state->get_game();
	
	//First, check time to shoot
	sint4 lastFired = unit->component("weapon")->get_int("last_fired");
	if (lastFired >= 0) {
		uint4 game_tick = game.tick();
		sint4 cooldown = unit->component("weapon")->get_int("cooldown");
		int timeToShoot = cooldown - (game_tick - (uint4)lastFired);

		if (timeToShoot > 0) {
			return false;
		}
	}

	//Get the target for this unit from the global coordinator
	//-1 = No target available.
	int targetId = coordinator->getTargetId(objId);
	GameObjPtr target = coordinator->getTargetObj(objId);
	if (targetId >= 0) {
		//ATTACK!
		Vector<sint4> args;
		args.push_back(targetId);
		unit->component("weapon")->set_action("attack", args);
		return true;
	}

	return false;
}

