#include "FireCoordinator.H"
#include "Blackboard.H"
#include "UnitAgent.H"

/*
	Coordinates the fire at enemy targets. It should be called each frame.
	The methods getTargetId() and getTargetObj() are used to get results from
	the coordination.
*/
void FireCoordinator::coordinate_fire(GameStateModulePtr mState, PFUpdaterPtr pfupdater) {
	//clock_t startTime = clock();

	//Clear the data structure
	data.clear();

	state = mState;
	mapCreator = pfupdater;

	const Game& game(state->get_game());

	//Step 1 - Get all possible fire combinations.
	const Game::ObjCont &objs = game.get_objs(mapCreator->playerId);

	FORALL (objs, it) {
		GameObjPtr gob = (*it)->get_GameObj();
		if (gob && gob->sod.in_game && *gob->sod.x >= 0 && !gob->is_dead()) {
			std::string objType = gob->bp_name();
			
			//int unitX = *gob->sod.x / TotalPF::CMTILE_WIDTH; TODO: Why / tile width????
			//int unitY = *gob->sod.y / TotalPF::CMTILE_WIDTH;
			int unitX = *gob->sod.x;
			int unitY = *gob->sod.y;

			if (objType == "tank" || objType == "marine") {
				if (canFire(unitX, unitY, gob, game)) {
					checkTargets(gob, game);
				}
			}
		}
	}
	
	//Uncomment for debugging of coordination
	/*if (data.hasEntries()) {
		cout << "Before processing" << endl;
		data.dump();
	}*/

	//Step 2 - Coordinate the fire to destroy as many enemy units as possible
	data.findAttackScheme();
	
	//Uncomment for debugging of coordination
	/*if (data.hasEntries()) {
		cout << "After processing" << endl;
		data.dump();
	}*/

	//Uncomment to measure execution time
	/*clock_t endTime = clock();
	double elapsed = ((double)((endTime - startTime) * 1000))/CLOCKS_PER_SEC;
	if (data.hasEntries()) {
		cout << "Iteration time: " << elapsed << endl;
	}*/
}

/*
	Returns the object Id for the target the specified unit shall fire at. Returns
	-1 if no target is available.
*/
int FireCoordinator::getTargetId(uint4 unitId) {
	return data.getTargetId(unitId);
}

/*
	Returns the GameObj for the target the specified unit shall fire at. Returns
	null if no target is available.
*/
GameObj* FireCoordinator::getTargetObj(uint4 unitId) {
	return data.getTargetObj(unitId);
}

/*
	Returns true if the specified unit can fire this frame.
	Two checks are made:
	- If the units weapon are in a cooldown period firing isn't possible
	- If the enemy unit chargemap has a 0 value at the units location no targets are nearby
*/
bool FireCoordinator::canFire(int unitX, int unitY, GameObjPtr unit, const Game& game) {
	//Step 1 - Check cooldown
	sint4 lastFired = unit->component("weapon")->get_int("last_fired");
	if (lastFired >= 0) {
		uint4 game_tick = game.tick();
		sint4 cooldown = unit->component("weapon")->get_int("cooldown");
		int timeToShoot = cooldown - (game_tick - (uint4)lastFired);

		if (timeToShoot > 0) {
			//Cooldown period. Not allowed to shoot
			return false;
		}
	}

	//Allowed to shoot. Proceed to check for enemies in range

	return true;
}

/*
	Checks for possible targets for this unit. Returns number of available targets.
	Possible targets are added to FCStructure.
*/
int FireCoordinator::checkTargets(GameObjPtr unit, const Game& game) {
	PlayerInfo& pi = game.get_cplayer_info();
	uint4 unitId = pi.get_id(unit);

	int found_cnt = 0;

	//Get fire range
	real8 fireRange = UnitAgent::TANK_FIRERANGE;
	if (Blackboard::getType(unit) == Blackboard::TYPE_MARINE) {
		fireRange = UnitAgent::MARINE_FIRERANGE;
	}

	//Iterate through enemy objects
	const Game::ObjCont &objs = game.get_objs(mapCreator->enemyId);
	FORALL (objs, it) {
		GameObj *gob = (*it)->get_GameObj();
		if (gob && gob->sod.in_game && *gob->sod.x >= 0 && !gob->is_dead()) {

			if (canAttackTarget(unit, gob, fireRange)) {
				found_cnt++;
				uint4 targetId = pi.get_id(gob);
				data.addAttackPossibility(unitId, unit, targetId, gob);
			}
		}
	}

	return found_cnt;
}

/*
	Checks if a unit can attack the specified target. Four checks are made:
	- Check if the Manhattan distance / 2 to the unit is less than firerange.
	- Check if unit shall attack the target based on global mode (i.e. if global mode is hunt units enemy bases are not attacked).
	- Check if the actual distance to the unit is less than firerange.
	- Check line of sight to target.

	The Manhattan distance check is used since it is a very fast way to decide if
	the rest of the checks are necessary or if the target is too far away.
*/
bool FireCoordinator::canAttackTarget(GameObjPtr unit, GameObjPtr destObj, real8 fireRange) {
	std::string objType = destObj->bp_name();
	
	//Step 1 - Check Manhattan distance
	real8 md = calculateMD(unit, destObj) / (real8)2;
	if (md > fireRange) {
		/*if (objType == "barracks") {
			cout << "failed: " << md << ">" << fireRange << endl;
		}*/

		//Outside Manhattan distance / 2
		return false;
	}
	
	//Step 2 - Check if we shall attack this type of unit
	if (objType == "tank" || objType == "marine" || objType == "worker" || objType == "controlCenter" || objType == "barracks" || objType == "factory") {
		//Attack!
	}
	else {
		//No object we can attack
		return false;
	}
	
	//Step 3 - Calculate real (euclidean) distance
	real8 ed = unit->distance_to(*destObj);
	if (ed > fireRange) {
		//Outside real distance
		return false;
	}
	
	//Step 4 - Check visibility
	Game& game = state->get_game();
	bool visible = game.can_see_obj(*unit, *destObj);
	if (!visible) {
		//Target is not visible
		return false;
	}
	
	//All checks passed
	return true;
}
/*
	Calculates the Manhattan distance between two game objects.
*/
real8 FireCoordinator::calculateMD(GameObjPtr unit, GameObjPtr destObj) {
	real8 x1 = (real8)(*unit->sod.x);
	real8 y1 = (real8)(*unit->sod.y);
	real8 x2 = (real8)(*destObj->sod.x);
	real8 y2 = (real8)(*destObj->sod.y);

	return abs(x1 - x2) + abs(y1 - y2);
}
