#ifndef FireCoordinator_H
#define FireCoordinator_H

#include "TotalPF.H"
#include "PFUpdater.H"
#include "GameObj.H"
#include "Game.H"
#include "GameStateModule.H"
#include "FCStructure.H"

/*
	This is the main class for the global coordinator of fire at enemy targets.

	The coordinator strives to destroy as many targets as possible each frame. If it isn't possible
	to completely destroy a target, weaker (damaged) targets and targets many units can fire at are
	preferred.
*/
class FireCoordinator {
	
	typedef TotalPF* TotalPFPtr;
	typedef GameObj* GameObjPtr;
	typedef PFUpdater* PFUpdaterPtr;
	typedef GameStateModule* GameStateModulePtr;

	public:

		/*
			Coordinates the fire at enemy targets. It should be called each frame.
			The methods getTargetId() and getTargetObj() are used to get results from
			the coordination.
		*/
		void coordinate_fire(GameStateModulePtr mState, PFUpdaterPtr pfupdater);

		/*
			Returns the object Id for the target the specified unit shall fire at. Returns
			-1 if no target is available.
		*/
		int getTargetId(uint4 unitId);

		/*
			Returns the GameObj for the target the specified unit shall fire at. Returns
			null if no target is available.
		*/
		GameObj* getTargetObj(uint4 unitId);

		/*
			Returns true if the specified unit can fire this frame.
			Two checks are made:
			- If the units weapon are in a cooldown period firing isn't possible
			- If the enemy unit chargemap has a 0 value at the units location no targets are nearby
		*/
		bool canFire(int unitX, int unitY, GameObjPtr unit, const Game& game);

		/*
			Checks for possible targets for this unit. Returns number of available targets.
			Possible targets are added to FCStructure.
		*/
		int checkTargets(GameObjPtr unit, const Game& game);

		/*
			Checks if a unit can attack the specified target. Four checks are made:
			- Check if the Manhattan distance / 2 to the unit is less than firerange.
			- Check if unit shall attack the target based on global mode (i.e. if global mode is hunt units enemy bases are not attacked).
			- Check if the actual distance to the unit is less than firerange.
			- Check line of sight to target.

			The Manhattan distance check is used since it is a very fast way to decide if
			the rest of the checks are necessary or if the target is too far away.
		*/
		bool canAttackTarget(GameObjPtr unit, GameObjPtr destObj, real8 fireRange);

		/*
			Calculates the Manhattan distance between two game objects.
		*/
		real8 calculateMD(GameObjPtr unit, GameObjPtr destObj);

	private:
		GameStateModulePtr state;
		PFUpdaterPtr mapCreator;
		FCStructure data;
};

#endif
