#ifndef PF_H
#define PF_H

#include "Trail.H"
#include "GameObj.H"

using namespace std;

/*
	Always FogOfWar for player 0 hack
	/libs/kernel/src/View.C ,line 393
*/

/*
	Struct used for explored map.
*/
struct ExplTile {
	int visitCnt;

	void visited() {
		visitCnt = 0;
	}
};

/*
	Struct for holding a mineral. 
	Note that assignedWorker is currently not in use.
*/
struct MineralPos {
	uint4 objId;
	uint4 assignedWorker;

	int x;
	int y;

	bool destroyed;

	bool isThisObj(uint4 mObjId) {
		if (objId == mObjId) {
			return true;
		}
		return false;
	}

	bool isAssignedTo(uint4 mObjId) {
		if (assignedWorker == mObjId) {
			return true;
		}
		return false;
	}
};

/*
	Struct for holding a dynamic (moveable) object.
*/
struct ObjPos {
	uint4 objId;
	sint4 type;

	int x;
	int y;
	int gametick;
	bool destroyed;

	bool isThisObj(uint4 mObjId) {
		if (objId == mObjId) {
			return true;
		}
		return false;
	}

	void update(uint4 mGametick, int mX, int mY) {
		destroyed = false;
		x = mX;
		y = mY;
		gametick = mGametick;
	}

	bool isMarine() {
		if (type == 1) return true;
		return false;
	}

	bool isTank() {
		if (type == 0) return true;
		return false;
	}

	bool isWorker() {
		if (type == 3) return true;
		return false;
	}
};

/*
	Struct for holding a base or other structure.
*/
struct BasePos {
	uint4 objId;
	GameObj* obj;
	
	int type;

	int x;
	int y;
	int x1;
	int x2;
	int y1;
	int y2;

	bool isThisObj(uint4 mObjId) {
		if (objId == mObjId) {
			return true;
		}
		return false;
	}

	bool isPositionInside(int cX, int cY) {
		return isPositionInside(cX, cY, 2);
	}

	bool isPositionInside(int cX, int cY, int marg) {
		int mMarg = 3 + marg;

		if (cX >= x1 - mMarg && cX <= x2 + mMarg) {
			if (cY >= y1 - mMarg && cY <= y2 + mMarg) {
				return true;
			}
		}
		return false;
	}

	bool isControlCenter() {
		if (type == 2) return true;
		return false;
	}

	bool isBarracks() {
		if (type == 4) return true;
		return false;
	}

	bool isFactory() {
		if (type == 5) return true;
		return false;
	}
};

/*
	Struct for holding an own worker or unit where the next position of the object is known.
*/
struct ObjFuturePos {
	uint4 objId;
	GameObj* obj;
	sint4 type;
	
	int x;
	int y;
	int nextX;
	int nextY;
	int gametick;
	bool destroyed;

	Trail trail;

	sint4 currentState;
	int miscState;
	int destX;
	int destY;

	bool isThisObj(uint4 mObjId) {
		if (objId == mObjId) {
			return true;
		}
		return false;
	}

	void update(uint4 mGametick, int mX, int mY) {
		destroyed = false;
		x = mX;
		y = mY;
		nextX = -1;
		nextY = -1;
		gametick = mGametick;
	}

	bool hasNextPos() {
		if (nextX == -1) return false;
		if (nextY == -1) return false;
		return true;
	}

	bool isMarine() {
		if (type == 1) return true;
		return false;
	}

	bool isTank() {
		if (type == 0) return true;
		return false;
	}

	bool isWorker() {
		if (type == 3) return true;
		return false;
	}

	bool isAvailable() {
		//A worker is available if it moves to or from mine spot
		if (currentState == 21 || currentState == 23) {
			return true;
		}
		return false;
	}
};

/*
	Holds a position.
*/
struct Position {
	int x;
	int y;

	bool found() {
		if (x == -1) return false;
		if (y == -1) return false;
		return true;
	}
};

/*
	This class holds lists of game objects, and has methods for calculating the potential objects generates in a point.
*/
class PF {
	
	typedef GameObj* GameObjPtr;
	
	public:

		/*
			Constructor.
		*/
		PF();

		/*
			Returns number of own units + workers.
		*/
		int getOU();

		/*
			Returns number of enemy units + workers.
		*/
		int getEU();

		/*
			Returns the number of own bases + structures.
		*/
		int getOB();

		/*
			Returns the number of enemy bases + structures.
		*/
		int getEB();

		/*
			Sets all own and enemy units as destroyed. Shall be called before updating the unit positions
			each frame. A unit which has its position updated are set as not destroyed, and after the
			position updates all units which still has the destroyed flag set can be removed.
		*/
		void setAllDestroyed();

		/*
			Removes destroyed own and enemy units.
			Must be used with the setAllDestroyed() method.
		*/
		void cleanUp();

		/*
			Updates the explored map. This map is used for explorers to find spots that needs
			to be explored.
		*/
		void updateExplMap(const Game& game);

		/*
			Returns the next point to visit for an explorer currently at position (unitX,unitY).
		*/
		Position getNextToVisit(int unitX, int unitY);

		/*
			Updates the position for mineral [objId].
		*/
		void updateMineralPosition(uint4 objId, int x, int y);

		/*
			Updates the position for sheep [objId].
		*/
		void updateSheepPosition(uint4 game_tick, uint4 objId, int x, int y);

		/*
			Updates the position for enemy unit [objId].
		*/
		void updateEnemyUnitPosition(uint4 game_tick, uint4 objId, sint4 unitType, int x, int y);

		/*
			Updates the position for own unit [objId].
		*/
		void updateUnitPosition(uint4 game_tick, GameObj* obj, uint4 objId, sint4 unitType, int x, int y);

		/*
			Sets the position the unit [objId] is moving to.
		*/
		void setNextUnitPosition(uint4 objId, int nextX, int nextY);
		
		/*
			Returns the potential generated by minerals in the point (pX,pY).
		*/
		real8 getMineralCharge(int pX, int pY, uint4 objId, int unitType, int marg, bool attractive);

		/*
			Returns the potential generated by sheep in the point (pX,pY).
		*/
		real8 getSheepCharge(int pX, int pY, uint4 game_tick, int baseRadius);

		/*
			Returns the potential generated by enemy units in the point (pX,pY).
		*/
		real8 getEnemyUnitCharge(int pX, int pY, bool offensive, int unitType, uint4 game_tick);

		/*
			Returns the potential generated by own units in the point (pX,pY).
		*/
		real8 getUnitCharge(int pX, int pY, uint4 objId, int unitType, uint4 game_tick);

		/*
			Removes all elements in the own- and enemy bases vectors.
		*/
		void clearBaseLists();

		/*
			Adds the base [objId] to a bases vector.
		*/
		void addBase(uint4 objId, GameObj* obj, int x, int y, int x1, int y1, int x2, int y2, bool own, int type);

		/*
			Removes a destroyed base previously located at (x,y) from the bases list.
		*/
		void removeBase(int x, int y);

		/*
			Dumps info about known bases to the console.
		*/
		void dumpBases();

		/*
			Returns the potential generated by own bases in the point (pX,pY).
		*/
		real8 getBaseCharge(int pX, int pY, int radius, bool attractive);

		/*
			Returns the potential generated by own bases when units are defending the own base.
		*/
		real8 getDefendBaseCharge(int pX, int pY, int unitType);

		/*
			Returns the potential generated by enemy bases in the point (pX,pY).
		*/
		real8 getEnemyBaseCharge(int pX, int pY, int unitType, bool attack);

		/*
			Returns the potential bases generates when finding buildspots.
		*/
		real8 getBuilderBaseCharge(int pX, int pY, uint4 objId, uint4 game_tick, bool useTrail);

		/*
			Returns the potential map edges generates when finding buildspots.
		*/
		real8 getBuilderEdgeCharge(int pX, int pY);

		//Worker FSM state
		sint4 getCurrentState(uint4 objId);

		/*
			Changes the current FSM state for a worker.
		*/
		void updateCurrentState(uint4 objId, sint4 newState);

		/*
			Returns the destination (if set) for a worker.
		*/
		Position getDestination(uint4 objId);

		/*
			Clears/removes the destination for a worker.
		*/
		void clearDestination(uint4 objId);

		/*
			Sets the destination for a worker.
		*/
		void setDestination(uint4 objId, Position pos);

		//Returns the different game object lists.
		vector<ObjFuturePos>* getUnitsList();
		vector<BasePos>* getBasesList();
		vector<ObjPos>* getEnemyUnitsList();
		vector<BasePos>* getEnemyBasesList();
		vector<MineralPos>* getMineralsList();
		MineralPos closestMineral; //temp var

		//Vectors holding the game objects.
		vector<ObjPos> sheep;
		vector<ObjPos> enemies;
		vector<ObjFuturePos> units;
		vector<BasePos> bases;
		vector<BasePos> enemyBases;
		vector<MineralPos> minerals;

	private:
	
		/*
			Returns the potential a mineral at (objX,objY) generates in the point (x,y).
		*/
		real8 calcMineralCharge(int x, int y, int objX, int objY, bool attractive, int baseRadius);

		/*
			Returns the potential a sheep at (objX,objY) generates in the point (x,y).
		*/
		real8 calcSheepCharge(int x, int y, int objX, int objY, int baseRadius);

		/*
			Returns the potential a marine at (objX,objY) generates in the point (x,y).
		*/
		real8 calcOffensiveMarineCharge(int x, int y, int objX, int objY);
		
		/*
			Returns the potential a marine at (objX,objY) generates in the point (x,y).
		*/
		real8 calcDefensiveMarineCharge(int x, int y, int objX, int objY);

		/*
			Returns the potential a tank at (objX,objY) generates in the point (x,y).
		*/
		real8 calcOffensiveTankCharge(int x, int y, int objX, int objY);
		
		/*
			Returns the potential a tank at (objX,objY) generates in the point (x,y).
		*/
		real8 calcDefensiveTankCharge(int x, int y, int objX, int objY);

		/*
			Returns the potential an own marine at (objX,objY) generates in the point (x,y).
		*/
		real8 calcMarineUnitCharge(int x, int y, int objX, int objY, int baseRadius);

		/*
			Returns the potential an own worker at (objX,objY) generates in the point (x,y).
		*/
		real8 calcWorkerUnitCharge(int x, int y, int objX, int objY, int baseRadius);

		/*
			Returns the potential an own tank at (objX,objY) generates in the point (x,y).
		*/
		real8 calcTankUnitCharge(int x, int y, int objX, int objY, int baseRadius);

		real8 calcUnitCharge(real8 dist, int radius);

		/*
			Returns the potential an own base at (objX,objY) generates in the point (x,y).
		*/
		real8 calcBaseCharge(int x, int y, int objX, int objY, bool inside, bool attractive);

		/*
			Returns the potential an enemy base at (objX,objY) generates in the point (x,y).
		*/
		real8 calcEnemyBaseCharge(int x, int y, int objX, int objY, bool inside, int unitType);

		/*
			Calculates the potential own bases and structures generates in point (x,y) when
			searching for a buildspot.
		*/
		real8 calcBuilderBaseCharge(int x, int y, int objX, int objY, bool inside);

		/*
			Calculates the potential enemy bases and structures generates in point (x,y) when
			searching for a buildspot.
		*/
		real8 calcBuilderEnemyBaseCharge(int x, int y, int objX, int objY, bool inside);

		/*
			Calculates the potential minerals generates in point (x,y) when
			searching for a buildspot.
		*/
		real8 calcBuilderMineralCharge(int x, int y, int objX, int objY);

		/*
			Sets that a tile at point (x,y) are visited. Range specifies the number of adjacent
			tiles that are also set as visited. Range 1 sets 3x3 tiles visited, range 2 5x5 tiles etc.
		*/
		void setTileVisited(int x, int y, int range);

		//Holds the explorer map
		ExplTile expl_map[64][64];

};

#endif
